from typing import TYPE_CHECKING
from sqlmodel import Session

from synda.database import engine
from synda.model.run import Run, RunStatus
from synda.utils.env import is_debug_enabled

if TYPE_CHECKING:
    from synda.config import Config


class Pipeline:
    def __init__(self, config: "Config"):
        self.config = config
        self.session = Session(engine)
        self.input_loader = config.input.get_loader()
        self.output_saver = config.output.get_saver()
        self.pipeline = config.pipeline
        self.run = Run.create_with_steps(self.session, config)

    def execute(self):
        try:
            input_nodes = self.input_loader.load(self.session)

            for step in self.run.steps:
                if is_debug_enabled():
                    print(step)

                executor = step.get_step_config().get_executor(
                    self.session, self.run, step
                )
                input_nodes = executor.execute_and_update_step(input_nodes)

            self.output_saver.save(input_nodes)

            self.run.update(self.session, RunStatus.FINISHED)
        except Exception as e:
            self.run.update(self.session, RunStatus.ERRORED)
            raise e



from abc import abstractmethod

from sqlmodel import Session

from synda.model.run import Run
from synda.model.step import Step, StepStatus
from synda.model.node import Node


class Executor:
    def __init__(self, session: Session, run: Run, step_model: Step):
        self.session = session
        self.run = run
        self.step_model = step_model
        self.config = step_model.get_step_config()

    def execute_and_update_step(self, input_nodes: list[Node]) -> list[Node]:
        try:
            self.step_model.set_running(self.session, input_nodes)

            output_nodes = self.execute(input_nodes)

            self.step_model.set_completed(self.session, input_nodes, output_nodes)

            return output_nodes
        except Exception as e:
            self.step_model.set_status(self.session, StepStatus.ERRORED)
            raise e

    @abstractmethod
    def execute(self, input_data: list[Node]):
        pass



from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING

from sqlmodel import Column, Relationship, SQLModel, Session, Field, JSON

from synda.database import engine
from synda.model.step import Step, StepStatus

if TYPE_CHECKING:
    from synda.config import Config


class RunStatus(str, Enum):
    RUNNING = "running"
    FINISHED = "finished"
    ERRORED = "errored"


class Run(SQLModel, table=True):
    id: int | None = Field(default=None, primary_key=True)
    status: RunStatus = Field(default=RunStatus.RUNNING, index=True)
    config: "Config" = Field(default=None, sa_column=Column(JSON))
    created_at: datetime = Field(default_factory=datetime.now)

    steps: list[Step] = Relationship(back_populates="run")

    @staticmethod
    def create_with_steps(session: Session, config: "Config") -> "Run":
        run = Run(config=config.model_dump())

        session.add(run)
        session.flush()

        steps = [
            Step(
                run_id=run.id,
                position=position,
                type=pipeline_step.type,
                method=pipeline_step.method,
                name=pipeline_step.name,
                step_config=pipeline_step.model_dump(),
                status=StepStatus.PENDING,
                run_at=datetime.now(),
            )
            for position, pipeline_step in enumerate(config.pipeline, start=1)
        ]

        session.add_all(steps)
        session.commit()
        session.refresh(run, ["steps"])

        return run

    def update(self, session: Session, status: RunStatus) -> "Run":
        self.status = status

        session.add(self)
        session.commit()
        session.refresh(self)
        return self

    def get_config(self) -> "Config":
        return Config.model_validate(self.config)



from typing import Annotated, Literal, Optional

from pydantic import BaseModel, Field, TypeAdapter
from sqlmodel import Session

from synda.config.step import Step
from synda.model.run import Run
from synda.model.step import Step as StepModel
from synda.pipeline.executor import Executor


class DeduplicateParametersTFIDF(BaseModel):
    strategy: Literal["exact", "fuzzy"] = Field(
        default="exact", description="Strategy for removing duplicates"
    )
    similarity_threshold: Optional[float] = Field(
        default=0.9, description="Threshold for similarity"
    )
    keep: Literal["first", "last"] = Field(
        default="first", description="Keep the first or last duplicate"
    )


class DeduplicateTFIDF(Step):
    type: str = "clean"
    method: Literal["deduplicate-tf-idf"]
    parameters: DeduplicateParametersTFIDF

    def get_executor(
        self, session: Session, run: Run, step_model: StepModel
    ) -> Executor:
        if self.method == "deduplicate-tf-idf":
            from synda.pipeline.clean import DeduplicateTFIDF

            return DeduplicateTFIDF(session, run, step_model)


Deduplicate = Annotated[DeduplicateTFIDF, Field(discriminator="method")]

deduplicate_adapter = TypeAdapter(Deduplicate)



from typing import Literal

from pydantic import BaseModel

from synda.config.input.csv import CSVInputProperties
from synda.config.input.xls import XLSInputProperties
from synda.config.input.database import DatabaseInputProperties

from synda.pipeline.input import InputLoader


class Input(BaseModel):
    type: Literal["csv", "xls"]
    properties: CSVInputProperties | XLSInputProperties  # | DatabaseSourceProperties

    def get_loader(self) -> InputLoader:
        if self.type == "csv":
            from synda.pipeline.input.csv_input_loader import CSVInputLoader

            return CSVInputLoader(self)
        elif self.type == "xls":
            from synda.pipeline.input.xls_input_loader import XLSInputLoader

            return XLSInputLoader(self)



from typing import Literal

from pydantic import BaseModel
from sqlmodel import Session

from synda.config.step import Step
from synda.model.run import Run
from synda.model.step import Step as StepModel


class GenerationParameters(BaseModel):
    provider: str = "openai"
    model: str = "gpt-4o-mini"
    template: str


class Generation(Step):
    type: str = "generation"
    method: Literal["llm"]
    parameters: GenerationParameters

    def get_executor(self, session: Session, run: Run, step_model: StepModel):
        if self.method == "llm":
            from synda.pipeline.generation import LLM

            return LLM(session, run, step_model)



import re

from sqlmodel import Session

from synda.model.node import Node


class PromptBuilder:
    @staticmethod
    def extract_template_variables(template: str) -> list[str]:
        pattern = r"{([^}]+)}"
        matches = re.finditer(pattern, template)
        return [match.group(1) for match in matches]

    @staticmethod
    def build(session: Session, template: str, input_data: list[Node]) -> list[str]:
        variables = PromptBuilder.extract_template_variables(template)
        prompts = []

        if len(variables) == 0:
            return [template for _ in input_data]

        parent_node_ids = set()
        for node in input_data:
            for variable_name in variables:
                parent_node_ids.add(node.ancestors[variable_name])

        parent_nodes = Node.get(session, list(parent_node_ids))

        for node in input_data:
            variable_value = {}
            for variable_name in variables:
                parent_node_id = node.ancestors[variable_name]
                parent_node = next(
                    node for node in parent_nodes if node.id == parent_node_id
                )
                variable_value[variable_name] = parent_node.value

            prompts.append(template.format(**variable_value))

        return prompts



from pathlib import Path
import typer

from synda.config import Config
from synda.pipeline import Pipeline


def generate_command(
    config_file: Path = typer.Argument(
        ...,
        help="Path to YAML configuration file",
        exists=True,
        file_okay=True,
        dir_okay=False,
        resolve_path=True,
    ),
):
    """Run a pipeline with provided configuration."""
    config = Config.load_config(config_file)
    pipeline = Pipeline(config)
    pipeline.execute()



import pandas as pd

from synda.config.output import Output
from synda.model.node import Node
from synda.pipeline.output.output_saver import OutputSaver


class XLSOutputSaver(OutputSaver):
    def __init__(self, output_config: Output):
        self.properties = output_config.properties
        super().__init__()

    def save(self, input_data: list[Node]) -> None:
        synthetic_data = [node.value for node in input_data]
        ablated_data = [node.is_ablated_text() for node in input_data]

        df = pd.DataFrame(
            {
                "synthetic data": synthetic_data,
                "ablated": ablated_data,
            }
        )

        df.to_excel(
            self.properties.path, sheet_name=self.properties.sheet_name, index=False
        )



from abc import ABC, abstractmethod

from pydantic import BaseModel, ConfigDict, model_validator


class OutputProperties(BaseModel, ABC):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    @model_validator(mode="after")
    def validate_properties(self) -> "OutputProperties":
        pass



from typing import TYPE_CHECKING, Union
from sqlmodel import SQLModel, Field, Relationship, Column, JSON, Session, select

from synda.model.step_node import StepNode

if TYPE_CHECKING:
    from synda.model.step import Step


class Node(SQLModel, table=True):
    id: int | None = Field(default=None, primary_key=True)
    parent_node_id: int | None = None
    ablated: bool = False
    value: str
    ancestors: dict = Field(default_factory=dict, sa_column=Column(JSON))

    step_node_links: list["StepNode"] = Relationship(back_populates="node")

    input_step: list["Step"] = Relationship(
        back_populates="input_nodes",
        link_model=StepNode,
        sa_relationship_kwargs={
            "primaryjoin": "and_(Node.id == StepNode.node_id, StepNode.relationship_type == 'input')",
            "secondaryjoin": "Step.id == StepNode.step_id",
            "secondary": "step_node",
            "overlaps": "step,node,step_node_links",
        },
    )

    output_step: list["Step"] = Relationship(
        back_populates="output_nodes",
        link_model=StepNode,
        sa_relationship_kwargs={
            "primaryjoin": "and_(Node.id == StepNode.node_id, StepNode.relationship_type == 'output')",
            "secondaryjoin": "Step.id == StepNode.step_id",
            "secondary": "step_node",
            "overlaps": "input_step,node,step_node_links,step",
        },
    )

    @classmethod
    def get(
        cls, session: Session, node_ids: Union[int, list[int]]
    ) -> Union["Node", list["Node"]]:
        if isinstance(node_ids, int):
            node_ids = [node_ids]
            single_result = True
        else:
            single_result = False

        query = select(cls).where(cls.id.in_(node_ids))
        results = session.exec(query).all()

        return results[0] if single_result and results else results

    def is_ablated_text(self) -> str:
        return "yes" if self.ablated else "no"



import pandas as pd

from sqlmodel import Session

from synda.pipeline.input import InputLoader
from synda.config.input import Input
from synda.model.node import Node


class XLSInputLoader(InputLoader):
    def __init__(self, input_config: Input):
        self.properties = input_config.properties
        super().__init__()

    def load(self, session: Session) -> list[Node]:
        df = pd.read_excel(self.properties.path, sheet_name=self.properties.sheet_name)
        target_list = df[self.properties.target_column]
        result = []

        for value in target_list.values:
            result.append(Node(value=value))

        self.persist_nodes(session, result)

        return result



from typing import List
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.feature_extraction.text import TfidfVectorizer
from sqlmodel import Session

from synda.model.run import Run
from synda.pipeline.executor import Executor
from synda.model.node import Node
from synda.progress_manager import ProgressManager
from synda.model.step import Step


class DeduplicateTFIDF(Executor):
    def __init__(self, session: Session, run: Run, step_model: Step):
        super().__init__(session, run, step_model)
        self.progress = ProgressManager("CLEAN")

    def execute(self, input_data: List[Node]) -> List[Node]:
        strategy = self.config.parameters.strategy
        similarity_threshold = self.config.parameters.similarity_threshold
        keep = self.config.parameters.keep

        with self.progress.task("  Cleaning...", len(input_data)) as advance:
            data_list = [str(node.value) for node in input_data]

            if strategy == "exact":
                result_list = self._remove_exact_duplicates(data_list, keep, advance)
            elif strategy == "fuzzy":
                result_list = self._remove_fuzzy_duplicates(
                    data_list, similarity_threshold, keep, advance
                )

        return [Node(value=item) for item in result_list]

    @staticmethod
    def _remove_exact_duplicates(data: List[str], keep: str, advance) -> List[str]:
        seen_values = set()
        result = []

        if keep == "first":
            for item in data:
                if item not in seen_values:
                    seen_values.add(item)
                    result.append(item)
                advance()
        elif keep == "last":
            for item in reversed(data):
                if item not in seen_values:
                    seen_values.add(item)
                    result.insert(0, item)
                advance()

        return result

    def _remove_fuzzy_duplicates(
        self, data: List[str], similarity_threshold: float, keep: str, advance
    ) -> List[str]:
        vectorizer = TfidfVectorizer(strip_accents="unicode")
        tfidf_matrix = vectorizer.fit_transform(data)
        similarity_matrix = cosine_similarity(tfidf_matrix)
        to_keep = set(range(len(data)))

        for i in range(len(data)):
            if i not in to_keep:
                continue
            for j in range(i + 1, len(data)):
                if j not in to_keep:
                    continue
                if similarity_matrix[i, j] > similarity_threshold:
                    if keep == "first":
                        to_keep.discard(j)
                    elif keep == "last":
                        to_keep.discard(i)
                        break
            advance()

        return [data[i] for i in sorted(to_keep)]



from abc import abstractmethod

from sqlmodel import Session

from synda.model.node import Node


class InputLoader:
    @abstractmethod
    def load(self, session: Session) -> list[Node]:
        pass

    @staticmethod
    def persist_nodes(session: Session, nodes: list[Node]) -> list[Node]:
        for node in nodes:
            session.add(node)
        session.flush()

        for node in nodes:
            node.ancestors = {"source": node.id}
            session.add(node)
        session.commit()

        for node in nodes:
            session.refresh(node)

        return nodes



import json
from typing import Literal

from pydantic import BaseModel
from sqlmodel import Session

from synda.model.provider import Provider
from synda.model.run import Run
from synda.model.step import Step
from synda.pipeline.executor import Executor
from synda.model.node import Node
from synda.progress_manager import ProgressManager
from synda.utils.env import is_debug_enabled
from synda.utils.llm_provider import LLMProvider


class LLMJudgeCriterionBinaryAnswer(BaseModel):
    answer: Literal["YES", "NO"]

    def is_positive_answer(self) -> bool:
        return self.answer == "YES"


class LLMJudgeBinary(Executor):
    def __init__(self, session: Session, run: Run, step_model: Step):
        super().__init__(session, run, step_model)
        self.progress = ProgressManager("ABLATION")
        self.provider = Provider.get(self.config.parameters.provider)
        self.model = self.config.parameters.model

    def execute(self, input_data: list[Node]):
        criteria = self.config.parameters.criteria
        result = []

        with self.progress.task(
            "  Ablating...", len(input_data) * len(criteria)
        ) as advance_node:
            for node in input_data:
                judge_answers = []
                for criterion in criteria:
                    prompt = self._build_binary_judge_prompt(node.value, criterion)
                    judge_answer = LLMProvider.call(
                        self.provider.name,
                        self.model,
                        self.provider.api_key,
                        prompt,
                        LLMJudgeCriterionBinaryAnswer,
                        url=self.provider.api_url,
                        format="json"
                    )
                    try:
                        judge_answer = LLMJudgeCriterionBinaryAnswer(
                            **json.loads(judge_answer)
                        )
                    except Exception as e:
                        print(f"Model response format is malformed: {e}")

                    judge_answers.append(judge_answer)
                    advance_node()

                ablated = not self._check_consensus(judge_answers)
                result_node = Node(
                    parent_node_id=node.id, value=node.value, ablated=ablated
                )
                result.append(result_node)

            if is_debug_enabled():
                print(f"Answer: {node.value}")
                print(f"Criteria: {str(criteria)}")
                print(f"Consensus: {self.config.parameters.consensus}")
                print(f"Judge answers: {judge_answers}")
                print(f"Ablated: {result_node.is_ablated_text()}\n")

        return result

    def _check_consensus(
        self, judge_answers: list[LLMJudgeCriterionBinaryAnswer]
    ) -> bool:
        consensus = self.config.parameters.consensus.lower()
        positive_answers = sum(
            1 for answer in judge_answers if answer.is_positive_answer()
        )
        total_answers = len(judge_answers)

        if total_answers == 0:
            return False

        match consensus:
            case "all":
                return positive_answers == total_answers
            case "any":
                return positive_answers > 0
            case "majority":
                return positive_answers > total_answers / 2
            case _:
                raise ValueError(f"Unknown consensus: {consensus}")

    # @todo use prompt builder for criterion
    # @todo move the criterion into step parameters
    @staticmethod
    def _build_binary_judge_prompt(candidate: str, criterion: str) -> str:
        return (
            f"You are an expert judge tasked with evaluating synthetic text data.\n"
            f"You are evaluating synthetic data against a given criterion.\n"
            "You must answer ONLY with {\"answer\": \"YES\"} or {\"answer\": \"NO\"}.\n"
            "Output {\"answer\": \"YES\"} when the criterion is fulfilled.\n"
            "Output {\"answer\": \"NO\"} when the criterion is NOT fulfilled.\n"
            f"------\n"
            f"criterion: Is the candidate written in english?\n"
            f"candidate: Great Britain is a bit pretentious to call itself “Great”.\n"
            "{\"answer\": \"YES\"}\n"
            f"------\n"
            f"criterion: Does the text contain more than 10 words?\n"
            f"candidate: The cat sleeps.\n"
            "{\"answer\": \"NO\"}\n"
            f"------\n"
            f"criterion: Does the text talks about the sun?\n"
            f"candidate: Synda is a synthetic data library.\n"
            "{\"answer\": \"NO\"}\n"
            f"------\n"
            f"criterion: {criterion}\n"
            f"candidate: {candidate}\n"
        )



from sqlmodel import Session

from synda.model.run import Run
from synda.pipeline.executor import Executor
from synda.model.node import Node
from synda.progress_manager import ProgressManager
from synda.model.step import Step


class Separator(Executor):
    def __init__(self, session: Session, run: Run, step_model: Step):
        super().__init__(session, run, step_model)
        self.progress = ProgressManager("SPLIT")

    def execute(self, input_data: list[Node]):
        result = []
        separator = self.config.parameters.separator
        keep_separator = self.config.parameters.keep_separator

        with self.progress.task("Separating...", len(input_data)) as advance:
            for node in input_data:
                text = node.value
                start = 0
                pos = text.find(separator)

                while pos != -1:
                    chunk_end = pos + len(separator) if keep_separator else pos
                    chunk = text[start:chunk_end]

                    if chunk:
                        result.append(Node(parent_node_id=node.id, value=chunk))

                    start = pos + len(separator)
                    pos = text.find(separator, start)

                if start < len(text):
                    result.append(Node(parent_node_id=node.id, value=text[start:]))

                advance()

        return result



from sqlmodel import Session

from synda.model.provider import Provider
from synda.model.run import Run
from synda.model.step import Step
from synda.pipeline.executor import Executor
from synda.model.node import Node
from synda.utils.llm_provider import LLMProvider
from synda.utils.prompt_builder import PromptBuilder
from synda.progress_manager import ProgressManager


class LLM(Executor):
    def __init__(self, session: Session, run: Run, step_model: Step):
        super().__init__(session, run, step_model)
        self.progress = ProgressManager("GENERATION")
        self.provider = Provider.get(self.config.parameters.provider)
        self.model = self.config.parameters.model

    def execute(self, input_data: list[Node]):
        template = self.config.parameters.template
        prompts = PromptBuilder.build(self.session, template, input_data)
        result = []

        with self.progress.task("Generating...", len(input_data)) as advance:
            for node, prompt in zip(input_data, prompts):
                llm_answer = LLMProvider.call(
                    self.provider.name,
                    self.model,
                    self.provider.api_key,
                    prompt,
                    url=self.provider.api_url,
                )
                result.append(Node(parent_node_id=node.id, value=llm_answer))
                advance()

        return result



from enum import Enum

import typer
from sqlalchemy.exc import IntegrityError, NoResultFound

from synda.model.provider import Provider


class ProviderAction(str, Enum):
    ADD = "add"
    DELETE = "delete"
    UPDATE = "update"


def add_provider(name: str, api_key: str | None, api_url: str | None) -> None:
    try:
        Provider.create(name=name, api_key=api_key, api_url=api_url)
        typer.secho(f"Successfully added provider: {name}", fg=typer.colors.GREEN)
    except IntegrityError as e:
        typer.secho(f"Provider {name} already exists", fg=typer.colors.YELLOW)
        raise typer.Exit(1)


def delete_provider(name: str) -> None:
    try:
        provider = Provider.get(name)
        provider.delete()
        typer.secho(f"Successfully deleted provider: {name}", fg=typer.colors.GREEN)
    except NoResultFound as e:
        typer.secho(
            f"Provider {name} not found in database. "
            "Please add it using 'synda provider add <name> --api-key <key>'",
            fg=typer.colors.RED,
        )
        raise typer.Exit(1)


def update_provider(name: str, api_key: str, api_url: str) -> None:
    if api_key is None and api_url is None:
        typer.secho(
            "API key or API url is required for updating a provider",
            fg=typer.colors.RED,
        )
        raise typer.Exit(1)

    try:
        provider = Provider.get(name)
        if api_key is not None:
            provider.update(api_key=api_key)
        if api_url is not None:
            provider.update(api_url=api_url)
        typer.secho(f"Successfully updated provider: {name}", fg=typer.colors.GREEN)
    except NoResultFound as e:
        typer.secho(f"Provider {name} not found", fg=typer.colors.RED)
        raise typer.Exit(1)


def provider_command(
    action: ProviderAction = typer.Argument(
        ..., help="Action to perform: add, delete, or update"
    ),
    model_provider: str = typer.Argument(..., help="Model provider name"),
    api_key: str = typer.Option(
        None,
        "--api-key",
        "-k",
        help="API key for model provider",
    ),
    api_url: str = typer.Option(
        None, "--api-url", "-k", help="API url to call for model provider"
    ),
):
    action_handlers = {
        ProviderAction.ADD: lambda: add_provider(model_provider, api_key, api_url),
        ProviderAction.DELETE: lambda: delete_provider(model_provider),
        ProviderAction.UPDATE: lambda: update_provider(
            model_provider, api_key, api_url
        ),
    }

    action_handlers[action]()



from typing import Literal, TYPE_CHECKING

from sqlmodel import SQLModel, Field, Relationship

if TYPE_CHECKING:
    from synda.config.step import Step
    from synda.model.node import Node


class StepNode(SQLModel, table=True):
    __tablename__ = "step_node"

    step_id: int = Field(foreign_key="step.id", primary_key=True)
    node_id: int = Field(foreign_key="node.id", primary_key=True)
    relationship_type: str = Field(
        index=True
    )  # @todo add constraint for "input" or "output" only

    step: "Step" = Relationship(back_populates="step_node_links")
    node: "Node" = Relationship(back_populates="step_node_links")



from sqlmodel import SQLModel, create_engine, Session

DATABASE_URL = "sqlite:///database.db"

engine = create_engine(DATABASE_URL)


def get_session():
    with Session(engine) as session:
        yield session


def init_db():
    SQLModel.metadata.create_all(engine)



from typing import Literal

from pydantic import BaseModel
from sqlmodel import Session

from synda.config.step import Step
from synda.model.run import Run
from synda.model.step import Step as StepModel


class AblationParameters(BaseModel):
    provider: str = "openai"
    model: str = "gpt-4o-mini"
    consensus: Literal["all", "any", "majority"]
    criteria: list[str]


class Ablation(Step):
    type: str = "ablation"
    method: Literal["llm-judge-binary"]
    parameters: AblationParameters

    def get_executor(self, session: Session, run: Run, step_model: StepModel):
        if self.method == "llm-judge-binary":
            from synda.pipeline.ablation import LLMJudgeBinary

            return LLMJudgeBinary(session, run, step_model)



from typing import Literal

from pydantic import BaseModel

from synda.config.output.csv import CSVOutputProperties
from synda.config.output.xls import XLSOutputProperties

from synda.pipeline.output.output_saver import OutputSaver


class Output(BaseModel):
    type: Literal["csv", "xls"]
    properties: CSVOutputProperties | XLSOutputProperties

    def get_saver(self) -> OutputSaver:
        if self.type == "csv":
            from synda.pipeline.output.csv_output_saver import CSVOutputSaver

            return CSVOutputSaver(self)

        if self.type == "xls":
            from synda.pipeline.output.xls_output_saver import XLSOutputSaver

            return XLSOutputSaver(self)



from sqlmodel import SQLModel, Session, Field, select

from synda.database import engine


class Provider(SQLModel, table=True):
    id: int | None = Field(default=None, primary_key=True)
    name: str = Field(index=True, unique=True)
    api_key: str | None = Field(default=None, unique=True)
    api_url: str | None = Field(default=None, unique=True)

    @staticmethod
    def create(
        name: str, api_key: str | None = None, api_url: str | None = None
    ) -> "Provider":
        with Session(engine) as session:
            provider = Provider(name=name, api_key=api_key, api_url=api_url)
            session.add(provider)
            session.commit()
            session.refresh(provider)
            return provider

    def update(self, **kwargs) -> "Provider":
        with Session(engine) as session:
            for field, value in kwargs.items():
                setattr(self, field, value)

            session.add(self)
            session.commit()
            session.refresh(self)
            return self

    def delete(self) -> None:
        with Session(engine) as session:
            session.delete(self)
            session.commit()

    @staticmethod
    def get(name: str) -> "Provider":
        with Session(engine) as session:
            return session.exec(select(Provider).where(Provider.name == name)).one()



import typer
from dotenv import load_dotenv

from synda.database import init_db
from synda.cli.provider import provider_command
from synda.cli.generate import generate_command


app = typer.Typer(
    name="synda",
    help="A synthetic data generator pipeline CLI.",
    add_completion=False,
)

app.command("provider")(provider_command)
app.command("generate")(generate_command)


def main():
    load_dotenv()
    init_db()
    app()



from litellm import completion
from pydantic import BaseModel


class LLMProvider:
    @staticmethod
    def call(
        provider: str,
        model: str,
        api_key: str,
        prompt: str,
        response_format: BaseModel | None = None,
        url: str | None = None,
        format: str | None = None,
    ) -> str:
        provider = LLMProvider._resolve_provider(provider)
        response = completion(
            model=f"{provider}/{model}",
            messages=[{"content": prompt, "role": "user"}],
            api_key=api_key,
            api_base=url,
            response_format=response_format,
            format=format,
        )

        return response["choices"][0]["message"]["content"]

    @staticmethod
    def _resolve_provider(provider: str):
        if provider == "ollama":
            return "ollama_chat"
        return provider



import sys
import yaml
from pathlib import Path

import typer
from pydantic import BaseModel, model_validator
from sqlmodel import Session, select

from synda.config.input import Input
from synda.config.output import Output
from synda.config.ablation import Ablation
from synda.config.generation import Generation
from synda.config.split import Split
from synda.config.clean import Deduplicate
from synda.database import engine
from synda.model.provider import Provider


class MissingProviderError(Exception):
    pass


class Config(BaseModel):
    input: Input
    pipeline: list[Split | Generation | Ablation | Deduplicate]
    output: Output

    @staticmethod
    def load_config(config_path: Path) -> "Config":
        try:
            with open(config_path, "r", encoding="utf-8") as file:
                return Config.model_validate(yaml.safe_load(file))
        except yaml.YAMLError as e:
            print(f"Error in YAML file: {e}")
            sys.exit(1)
        except FileNotFoundError:
            print(f"The file {config_path} doesn't exist")
            sys.exit(1)
        except ValueError as e:
            print(f"Configuration validation error: {e}")
            sys.exit(1)
        except MissingProviderError as e:
            typer.secho(
                f"The following providers are not configured: {str(e)}.\n"
                "Please add them using 'synda provider add <name> --api-key <key>'",
                fg=typer.colors.RED,
            )
            raise typer.Exit(1)

    @model_validator(mode="after")
    def validate_providers(self) -> "Config":
        required_providers = set()
        for step in self.pipeline:
            if isinstance(step, (Generation, Ablation)):
                required_providers.add(step.parameters.provider)

        if not required_providers:
            return self

        with Session(engine) as session:
            statement = select(Provider).where(
                Provider.name.in_(required_providers)  # noqa
            )
            existing_providers = {provider.name for provider in session.exec(statement)}

            missing_providers = required_providers - existing_providers
            if missing_providers:
                providers_list = ", ".join(missing_providers)
                raise MissingProviderError(f"{providers_list}")

        return self



from typing import Annotated, Literal, Union

from pydantic import BaseModel, Field, TypeAdapter
from sqlmodel import Session

from synda.config.step import Step
from synda.model.run import Run
from synda.model.step import Step as StepModel
from synda.pipeline.executor import Executor


class ChunkParameters(BaseModel):
    size: int = Field(
        default=500, gt=0, lt=10000, description="Size of each chunk in characters"
    )


class ChunkSplit(Step):
    type: Literal["split"]
    method: Literal["chunk"]
    parameters: ChunkParameters

    def get_executor(
        self, session: Session, run: Run, step_model: StepModel
    ) -> Executor:
        from synda.pipeline.split import Chunk

        return Chunk(session, run, step_model)


class SeparatorParameters(BaseModel):
    separator: str = Field(default=".", description="The separator character(s)")
    keep_separator: bool = Field(
        default=True, description="Should keep the separator character(s)"
    )


class SeparatorSplit(Step):
    type: Literal["split"]
    method: Literal["separator"]
    parameters: SeparatorParameters

    def get_executor(
        self, session: Session, run: Run, step_model: StepModel
    ) -> Executor:
        from synda.pipeline.split import Separator

        return Separator(session, run, step_model)


Split = Annotated[Union[ChunkSplit, SeparatorSplit], Field(discriminator="method")]

split_adapter = TypeAdapter(Split)



from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING

from sqlmodel import Column, SQLModel, Field, Relationship, JSON, Session

from synda.model.step_node import StepNode
from synda.model.node import Node

if TYPE_CHECKING:
    from synda.config.step import Step as StepConfig
    from synda.model.run import Run


class StepStatus(Enum):
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    ERRORED = "errored"


class Step(SQLModel, table=True):
    id: int | None = Field(default=None, primary_key=True)
    run_id: int = Field(foreign_key="run.id")
    position: int = Field()
    type: str = Field(index=True)
    method: str = Field()
    name: str = Field(index=True)
    step_config: "StepConfig" = Field(default_factory=dict, sa_column=Column(JSON))
    status: StepStatus = Field(default=StepStatus.PENDING)
    run_at: datetime | None = Field()

    run: "Run" = Relationship(back_populates="steps")

    step_node_links: list["StepNode"] = Relationship(
        back_populates="step",
        sa_relationship_kwargs={"overlaps": "input_step,output_step"},
    )

    input_nodes: list[Node] = Relationship(
        back_populates="input_step",
        link_model=StepNode,
        sa_relationship_kwargs={
            "primaryjoin": "and_(Step.id == StepNode.step_id, StepNode.relationship_type == 'input')",
            "secondaryjoin": "Node.id == StepNode.node_id",
            "secondary": "step_node",
            "overlaps": "output_step,step,step_node_links,node",
        },
    )

    output_nodes: list[Node] = Relationship(
        back_populates="output_step",
        link_model=StepNode,
        sa_relationship_kwargs={
            "primaryjoin": "and_(Step.id == StepNode.step_id, StepNode.relationship_type == 'output')",
            "secondaryjoin": "Node.id == StepNode.node_id",
            "secondary": "step_node",
            "overlaps": "input_nodes,input_step,step,step_node_links,node",
        },
    )

    def set_status(self, session: Session, status: str) -> "Step":
        self.status = status
        session.add(self)
        session.commit()
        session.refresh(self)

        return self

    def set_running(self, session: Session, input_nodes: list[Node]) -> "Step":
        self.status = StepStatus.RUNNING
        self.run_at = datetime.now()

        for node in input_nodes:
            if node.id is None:
                session.add(node)
        session.flush()

        for node in input_nodes:
            step_node = StepNode(
                step_id=self.id, node_id=node.id, relationship_type="input"
            )
            session.add(step_node)

        session.add(self)
        session.commit()
        session.refresh(self)

        return self

    def set_completed(
        self, session: Session, input_nodes: list[Node], output_nodes: list[Node]
    ) -> "Step":
        self.status = StepStatus.COMPLETED

        self._create_nodes_with_ancestors(session, input_nodes, output_nodes)
        self._map_nodes_to_step(session, output_nodes)

        session.add(self)
        session.commit()
        session.refresh(self)

        return self

    def _create_nodes_with_ancestors(
        self, session: Session, input_nodes: list[Node], output_nodes: list[Node]
    ):
        for node in output_nodes:
            if node.id is None:
                session.add(node)
        session.flush()

        for node in output_nodes:
            parent_id = node.parent_node_id

            if parent_id is None:
                continue

            parent_node = next(node for node in input_nodes if node.id == parent_id)
            node.ancestors = parent_node.ancestors | {self.name: node.id}
            session.add(node)

    def _map_nodes_to_step(self, session: Session, output_nodes: list[Node]):
        for node in output_nodes:
            step_node = StepNode(
                step_id=self.id, node_id=node.id, relationship_type="output"
            )
            session.add(step_node)

    def get_step_config(self) -> "StepConfig":
        match self.type:
            case "split":
                from synda.config.split import split_adapter

                return split_adapter.validate_python(self.step_config)
            case "generation":
                from synda.config.generation import Generation

                return Generation.model_validate(self.step_config)
            case "ablation":
                from synda.config.ablation import Ablation

                return Ablation.model_validate(self.step_config)
            case "clean":
                from synda.config.clean import Deduplicate

                return Deduplicate.model_validate(self.step_config)
            case _:
                raise ValueError(f"Unknown step type: {self.type}")



from enum import Enum
from rich.progress import (
    Progress,
    TextColumn,
    BarColumn,
    TaskProgressColumn,
    SpinnerColumn,
    MofNCompleteColumn,
    TimeElapsedColumn,
)
from rich.console import Console
from contextlib import contextmanager


class ExecutorColor(Enum):
    SPLIT = "red"
    GENERATION = "green"
    ABLATION = "cyan"
    CLEAN = "blue"


class ProgressManager:
    def __init__(self, executor_type: str):
        self.console = Console()
        self.color = ExecutorColor[executor_type.upper()].value
        self.progress = Progress(
            SpinnerColumn(speed=2),
            TextColumn(f"[{self.color}]" + "{task.description}", justify="right"),
            BarColumn(),
            MofNCompleteColumn(),
            TaskProgressColumn(),
            TimeElapsedColumn(),
        )

    @contextmanager
    def task(self, description: str, total: int, transient: bool = False):
        with self.progress as progress:
            task_id = progress.add_task(description, total=total, transient=transient)
            yield lambda: progress.advance(task_id)



from sqlmodel import Session

from synda.model.run import Run
from synda.pipeline.executor import Executor
from synda.model.node import Node
from synda.progress_manager import ProgressManager
from synda.model.step import Step


class Chunk(Executor):
    def __init__(self, session: Session, run: Run, step_model: Step):
        super().__init__(session, run, step_model)
        self.progress = ProgressManager("SPLIT")

    def execute(self, input_data: list[Node]):
        result = []
        size = self.config.parameters.size

        with self.progress.task("  Chunking...", len(input_data)) as advance:
            for node in input_data:
                text = node.value

                while text:
                    chunk = text[:size]
                    text = text[size:]
                    result.append(Node(parent_node_id=node.id, value=chunk))

                advance()

        return result



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from pydantic import BaseModel, model_validator
from sqlmodel import Session

from synda.model.run import Run
from synda.model.step import Step

if TYPE_CHECKING:
    from synda.pipeline.executor import Executor


class Step(BaseModel, ABC):
    type: str
    method: str
    name: str | None = None
    parameters: dict

    @model_validator(mode="after")
    def set_default_name(self) -> str:
        if self.name is None:
            self.name = f"{self.type}_{self.method}"
        return self

    @abstractmethod
    def get_executor(self, session: Session, run: Run, step_model: Step) -> "Executor":
        pass



import os

import pandas as pd
from pydantic import Field, model_validator

from synda.config.input.input_properties import InputProperties


class XLSInputProperties(InputProperties):
    path: str
    target_column: str
    sheet_name: str = Field(default="Sheet1")

    @model_validator(mode="after")
    def validate_properties(self) -> "XLSInputProperties":
        self._validate_path()
        self._validate_file()
        return self

    def _validate_path(self) -> None:
        if not os.path.isfile(self.path):
            raise ValueError(f"Source file does not exist: {self.path}")

    def _validate_file(self) -> None:
        try:
            df = pd.read_excel(self.path, sheet_name=self.sheet_name)
            if self.target_column not in df.columns:
                raise ValueError(
                    f"Target column '{self.target_column}' not found in XLS file. Available columns: {', '.join(df.columns)}"
                )
        except pd.errors.EmptyDataError:
            raise ValueError(f"XLS file is empty: {self.path}")

        except pd.errors.ParserError:
            raise ValueError(f"Unable to parse XLS file: {self.path}")



import os

from pydantic import Field, model_validator

from synda.config.output.output_properties import OutputProperties


class CSVOutputProperties(OutputProperties):
    path: str
    separator: str = Field(default=";")

    @model_validator(mode="after")
    def validate_properties(self) -> "CSVOutputProperties":
        self._validate_path()
        return self

    def _validate_path(self) -> None:
        if os.path.isfile(self.path):
            raise ValueError(f"Output file already exist: {self.path}")
